"""Logging module."""
import logging
import sys
from pathlib import Path

from fastflows.config.app import configuration as cfg
from fastflows.schemas.logging import LoggerModel
from fastflows.schemas.logging import LoggingBase
from loguru import logger


class InterceptHandler(logging.Handler):
    """Custom logging interceptor."""

    loglevel_mapping = {
        50: "CRITICAL",
        40: "ERROR",
        30: "WARNING",
        20: "INFO",
        10: "DEBUG",
        0: "NOTSET",
    }

    def emit(self, record):
        """Emits a logging record."""
        try:
            level = logger.level(record.levelname).name
        except AttributeError:
            level = self.loglevel_mapping[record.levelno]

        frame, depth = logging.currentframe(), 2
        while frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1

        log = logger.bind(request_id="app")
        log.opt(depth=depth, exception=record.exc_info).log(
            level,
            record.getMessage(),
        )


class CustomizeLogger:
    """Handle logger customization."""

    @classmethod
    def make_logger(cls, config: LoggerModel):
        """Create a logger instance."""
        logging_config = config.logger

        logger = cls.customize_logging(
            filepath=logging_config.path,
            level=logging_config.level,
            enqueue=logging_config.enqueue,
            retention=logging_config.retention,
            rotation=logging_config.rotation,
            format=logging_config.format_,
        )
        return logger

    @classmethod
    def customize_logging(
        cls,
        filepath: Path,
        level: str,
        enqueue: bool,
        rotation: str,
        retention: str,
        format: str,
    ):
        """Customize logging configuration."""
        logger.remove()
        logger.add(
            sys.stdout,
            enqueue=enqueue,
            backtrace=True,
            level=level.upper(),
            format=format,
        )
        logger.add(
            str(filepath),
            rotation=rotation,
            retention=retention,
            enqueue=enqueue,
            backtrace=True,
            level=level.upper(),
            format=format,
        )
        logging.basicConfig(handlers=[InterceptHandler()], level=0)
        logging.getLogger("uvicorn.access").handlers = [InterceptHandler()]
        for _log in ["uvicorn", "uvicorn.error", "fastapi"]:
            _logger = logging.getLogger(_log)
            _logger.handlers = [InterceptHandler()]

        return logger.bind(request_id=None, method=None)


def create_logger(name: str):
    """Create a logger instance."""
    logger = logging.getLogger(name)
    config = LoggerModel(
        logger=LoggingBase(
            path=Path(cfg.LOG_PATH) / cfg.LOG_FILENAME,
            level=cfg.LOG_LEVEL,
            enqueue=cfg.LOG_ENQUEUE,
            retention=cfg.LOG_RETENTION,
            rotation=cfg.LOG_ROTATION,
            format_=cfg.LOG_FORMAT,
        ),
    )
    logger = CustomizeLogger.make_logger(config)

    return logger



"""Logging module."""
from pathlib import Path

from pydantic import BaseModel


class LoggingBase(BaseModel):
    """Base logging model."""

    path: Path
    level: str
    enqueue: bool
    retention: str
    rotation: str
    format_: str


class LoggerModel(BaseModel):
    """Logger model."""

    logger: LoggingBase



import typer
import sys
from typing import List
from rich import print as rprint
from functools import wraps
import traceback
from typing import Optional
from fastflows.config.app import configuration
from fastflows.schemas.flow import FlowDeployInput
from fastflows.schemas.flow_run import FlowRunInput, StateBase, FlowRunStateEnum
from fastflows.schemas.deployment import DeploymentInputParams
from fastflows.main import app_run
from fastflows.core.flow_run import (
    update_flow_run_state,
    get_flow_run_details,
)
from fastflows.core.flow import (
    run_flow,
    list_flows_from_file_home,
    deploy_flows,
    get_flow_runs_list,
)
from fastflows.core.task_run import update_task_run_state, get_task_run_state

app = typer.Typer()


def catch_exceptions(func):
    """decorator to remove traceback in command line & convert output to red beutiful error message,
    to disable option & see traceback: set env FASTFLOW_DEBUG = 1
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            result = func(*args, **kwargs)
            return result
        except Exception as e:
            if configuration.FASTFLOW_DEBUG == 1:
                typer.echo(traceback.format_exc())
            message = f"\n{e.args[0]}"
            typer.secho(message, fg=typer.colors.RED, err=True)
            sys.exit(1)

    return wrapper


@app.command()
@catch_exceptions
def server():
    """Start FastFlows server"""
    typer.echo("Starting FastFlows server")
    app_run()


# ---- Flow runs commands
flow_runs = typer.Typer()
app.add_typer(
    flow_runs, name="flow-runs", help="Operate with Flow Runs: get state, update state"
)


@flow_runs.command(name="list", help="List flow runs by flow_name or flow_id")
@catch_exceptions
def list_flow_runs(
    flow_name: str,
    flow_id: bool = typer.Option(False, help="List by flow_id"),
):
    typer.echo(
        f"Get list flow runs for flow {'with name' if not flow_id else 'with id'}: {flow_name}"
    )
    result = get_flow_runs_list(flow_name, flow_id)
    typer.echo(result)


@flow_runs.command(name="state", help="Get details for flow_run_id")
@catch_exceptions
def flow_run_state(flow_run_id: str):
    typer.echo(f"Get flow run state for flow_run_id {flow_run_id}")
    result = get_flow_run_details(flow_run_id)
    typer.echo(result)


@flow_runs.command(name="update", help="Update flow run state by flow_run_id")
@catch_exceptions
def update_flow_run(
    flow_run_id: str,
    state: str = typer.Option(
        FlowRunStateEnum.CANCELLED.value,
        help="State to set for new flow run - by default 'Scheduled' (mean will be runned imidiatly)",
    ),
):
    typer.echo(f"Set state {state} for flow_run_id {flow_run_id}")
    result = update_flow_run_state(flow_run_id, state=StateBase(type=state))
    typer.echo(result)


# ---- Flows commands
flows_app = typer.Typer()
app.add_typer(
    flows_app, name="flows", help="Operate with Flows: get state, update state"
)


def process_params_from_str(params: str) -> dict:
    params = params.strip().split()
    parameters = {}
    for pair in params:
        key, value = pair.split("=", 1)
        parameters[key] = value
    return parameters


@catch_exceptions
@flows_app.command(help="Run flow by Name or Id")
def run(
    flow_name: str,
    flow_id: bool = typer.Option(False, help="run by flow_id"),
    state: str = typer.Option(
        "PENDING",
        help="State to set for new flow run - by default 'Scheduled' (mean will be runned imidiatly)",
    ),
    params: Optional[str] = typer.Option(None, help="Parameters to pass to Flow Id"),
):
    typer.echo(f"Run flow: {flow_name}")
    if params:
        parameters = process_params_from_str(params)
    else:
        parameters = {}
    result = run_flow(
        flow_name,
        by_id=flow_id,
        flow_run_input=FlowRunInput(state={"type": state}, parameters=parameters),
    )
    typer.echo(f"Created flow run with id: {result.id}\n")
    typer.echo("Details:")
    rprint(result)


@flows_app.command()
@catch_exceptions
def list(flow_path: Optional[str] = configuration.FLOWS_HOME):
    """List all flows from FLOWS_HOME"""
    typer.echo("\nAll flows from FLOWS_HOME: \n")
    for flow in list_flows_from_file_home(flow_path):
        typer.echo(flow)
    typer.echo("\n")


@flows_app.command()
@catch_exceptions
def deploy(
    flows_home_path: str = typer.Argument(configuration.FLOWS_HOME),
    flow_name: Optional[str] = typer.Option(None, help="Flow name to deploy"),
    flow_path: Optional[str] = typer.Option(None, help="Flow path to deploy"),
    schedule: Optional[str] = typer.Option(
        None,
        help="""Schedule that should be used in the Deployment. If it will be used without `flow_path` or `flow_name` options
        this mean that ALL flows from path will be deployed with that schedule.""",
    ),
    active: Optional[bool] = typer.Option(True, help="Activate flow after deploy"),
    tags: Optional[List[str]] = typer.Option(None, help="Tags for deployment"),
    params: Optional[str] = typer.Option(None, help="Parameters for deployment"),
    force: bool = typer.Option(False, help="Force re-deploy all flows"),
):
    """Register flows in FastFlows & Prefect server"""
    if flow_name:
        if not flow_path:
            sub_message = f"from directory: {flows_home_path}"
        else:
            sub_message = f"from path: {flow_path}"
        typer.echo(f"Deploy flow '{flow_name}' {sub_message}")

    else:
        typer.echo(f"Deploy all flows from path: {flow_path}")

    deploy_flows(
        flow_input=FlowDeployInput(
            flows_home_path=flows_home_path,
            flow_name=flow_name,
            flow_path=flow_path,
            deployment_params=DeploymentInputParams(
                schedule=schedule,
                is_schedule_active=active,
                parameters=params,
                tags=tags,
            ),
            force=force,
        )
    )


# ---- Task Runs Comands
task_runs = typer.Typer()
app.add_typer(
    task_runs, name="task-runs", help="Operate with Tasks Runs: get state, update state"
)


@task_runs.command(name="update")
@catch_exceptions
def update_task_runs(task_run_id: str):
    typer.echo(f"Update task with id: {task_run_id}")
    update_task_run_state(task_run_id)


@task_runs.command(name="state")
@catch_exceptions
def task_run_state(task_run_id: str):
    typer.echo(f"Get state for task run task with id: {task_run_id}")
    get_task_run_state(task_run_id)



from fastflows.providers.prefect import PrefectProvider

# it's here because it is default right now, if will be supported more providers:
# add logic to get provider from configuration
provider = PrefectProvider()



"""App exceptions for frontend module."""
from app.config.logging import create_logger
from app.utils.app_exceptions import AppException


logger = create_logger(
    name="app.utils.get_list_of_app_exceptions_for_frontend",
)


logger.debug([e for e in dir(AppException) if "__" not in e])



from fastapi import APIRouter
from fastflows.schemas.flow_run import FlowRunState, StateBase, UpdateStateResponse
from fastflows.core.flow_run import (
    update_flow_run_state,
    get_flow_run_details,
)

router = APIRouter(prefix="/flow-runs", tags=["flows"])


@router.get("/{flow_run_id}", response_model=FlowRunState)
async def get_flow_run_details_route(flow_run_id: str):
    """
    :param flow_run_id: Flow Run Id in Prefect to get info about
    """
    return get_flow_run_details(flow_run_id)


@router.patch("/{flow_run_id}/state", response_model=UpdateStateResponse)
async def update_flow_run_state_route(
    flow_run_id: str, state: StateBase
) -> UpdateStateResponse:
    """
    :param flow_run_id: Flow Run Id in Prefect to update
    :body state: State to update (example: {'type': 'CANCELLED})
    """
    return update_flow_run_state(flow_run_id, state=state)



import datetime
from pydantic import BaseModel
from typing import Optional
from enum import Enum
import hashlib


class DefaultAPIResponseModel(BaseModel):
    status = "success"


class Schedule(BaseModel):

    interval: int
    timezone: str = "UTC"
    anchor_date: datetime.datetime


def get_hash_from_data(data: str) -> str:
    return hashlib.md5(data.encode()).hexdigest()


class Status(str, Enum):
    ACCEPT = "ACCEPT"
    ABORT = "ABORT"


class Details(BaseModel):
    type: str
    reason: Optional[str]



from functools import wraps
from fastflows.errors import FlowNotFound
from fastapi import HTTPException


def handle_rest_errors(func):
    @wraps(func)
    async def wrapper(*args, **kwargs):
        try:
            return await func(*args, **kwargs)
        except FlowNotFound as e:
            raise HTTPException(status_code=404, detail=str(e))

    return wrapper



import datetime
from enum import Enum
from pydantic import BaseModel, Field
from typing import Optional, List
from fastflows.schemas.deployment import FlowRunner
from fastflows.schemas.misc import Status, Details


class FlowRunStateEnum(str, Enum):

    SCHEDULED = "SCHEDULED"
    PENDING = "PENDING"
    RUNNING = "RUNNING"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"
    CANCELLED = "CANCELLED"


class StateDetails(BaseModel):
    flow_run_id: str
    task_run_id: Optional[str]
    child_flow_run_id: Optional[str]
    scheduled_time: Optional[datetime.datetime]
    cache_key: Optional[str]
    cache_expiration: Optional[str]


class StateBase(BaseModel):
    type: FlowRunStateEnum


class FlowRunInput(BaseModel):

    state: Optional[StateBase] = StateBase(type=FlowRunStateEnum.SCHEDULED)
    parameters: Optional[dict] = Field(default_factory=dict)


class State(StateBase):
    id: str
    name: str
    timestamp: datetime.datetime
    message: Optional[str]
    data: Optional[str]
    state_details: Optional[StateDetails]


class FlowRunState(BaseModel):

    id: str
    state: State


class FlowRunResponse(FlowRunState):
    name: str
    created: datetime.datetime
    updated: datetime.datetime
    name: str
    flow_id: str
    state_id: str
    deployment_id: Optional[str]
    flow_version: Optional[str]
    parameters: dict
    idempotency_key: Optional[str]
    context: dict
    empirical_policy: dict
    empirical_config: dict
    tags: List[str]
    parent_task_run_id: Optional[str]
    state_type: FlowRunStateEnum
    state_name: str
    run_count: int
    expected_start_time: Optional[datetime.datetime]
    next_scheduled_start_time: Optional[datetime.datetime]
    start_time: Optional[datetime.datetime]
    end_time: Optional[datetime.datetime]
    total_run_time: float
    estimated_run_time: float
    estimated_start_time_delta: float
    auto_scheduled: bool


class InitFlowRun(BaseModel):
    name: Optional[str]
    parameters: Optional[dict] = Field(default_factory=dict)
    idempotency_key: Optional[str]
    context: Optional[dict] = Field(default_factory=dict)
    tags: Optional[list] = Field(default_factory=list)
    flow_runner: Optional[FlowRunner]
    state: Optional[StateBase]


class UpdateStateResponse(BaseModel):
    # can be null if 'ABORT' status
    state: Optional[State]
    status: Status
    details: Details



import ast
import os
import sys
import pydantic
from fastflows.schemas.flow_data import (
    FlowDataFromFile,
    ScheduleFromFile,
    FlowData,
    TagsFromFile,
    Schedule,
)
from typing import List, Optional, Tuple
from fastflows.config.app import configuration as cfg
from fastflows.errors import FastFlowException


class FlowFileReader:
    def __init__(
        self, file_path: Optional[str] = None, flow_data: Optional[str] = None
    ) -> None:

        if not file_path and not flow_data:
            raise ValueError(
                "FlowFileReader expect one of 2 arguments: file_path or flow_data"
            )

        self.file_path = file_path
        self.file_data = flow_data or self._read_flow_file()
        # can be flow, but without name in decorator
        # todo: add test for both cases + case when there is no flow
        # can be several flows in one file
        self.flows: List[FlowDataFromFile] = []
        self._get_information_from_file()
        self.is_flows = bool(len(self.flows) > 0)

    def _read_flow_file(self) -> str:
        with open(self.file_path, "r") as f:
            return f.read()

    def _get_data_from_comment(self, key) -> Tuple[int, str]:
        for num, line in enumerate(self.file_data.split("\n")):
            line = line.strip()
            if line.startswith("#") and key in line:
                yield num, line.split(key)[1].replace(":", "", 1).strip()

    def _exctract_schedule_data(self):
        """expected comment in flow: `# schedule: interval=3600,anchor_date=2020-01-01T00:00:00Z,timezone=UTC`"""
        schedules = []
        key = cfg.SCHEDULE_PROPERTY
        for num, line in self._get_data_from_comment(key):
            if line:
                data = [item.strip() for item in line.split(",")]
                schedule_data = {}
                for item in data:
                    item = item.split("=")
                    if item[0] in ScheduleFromFile.__fields__:
                        schedule_data[item[0]] = item[1]
                try:
                    schedules.append(ScheduleFromFile(lineno=num, **schedule_data))
                except pydantic.error_wrappers.ValidationError:
                    raise FastFlowException(
                        "Wrong schedule format."
                        "Schedule in Flow File should be defined as comment line with interval & anchor_date & timezone"
                        "Example: `# schedule: interval=3600,anchor_date=2020-01-01T00:00:00Z,timezone=UTC`"
                    )

        return schedules

    def _exctract_tags(self):
        """expected comment in flow: `# tags: data_flow, some_tag`"""
        exctracted_tags = []
        key = cfg.TAGS_PROPERTY
        for num, line in self._get_data_from_comment(key):
            if line:
                tags = [item.strip() for item in line.split(",")]
                if tags:
                    exctracted_tags.append(TagsFromFile(lineno=num, tags=tags))
        return exctracted_tags

    def _get_information_from_file(self) -> None:
        parsed_ast = ast.parse(self.file_data)
        schedules: ScheduleFromFile = self._exctract_schedule_data()
        tags: TagsFromFile = self._exctract_tags()
        functions_in_file = [
            node
            for node in ast.walk(parsed_ast)
            if isinstance(node, ast.FunctionDef) and node.decorator_list
        ]

        # schedule should be written in comment with # before flow defenition

        for function in functions_in_file:
            self._process_function_node(function, schedules, tags)

    def _process_function_node(
        self,
        function: ast.FunctionDef,
        schedules: List[ScheduleFromFile],
        tags: List[TagsFromFile],
    ) -> None:
        decor = function.decorator_list[0]
        flow_name = None
        is_flow = False
        if isinstance(decor, ast.Name) and decor.id == "flow":
            # mean no call, no args, no flow name, but it is a flow
            is_flow = True
        elif isinstance(decor, ast.Call) and decor.func.id == "flow":
            # mean it is call, so we have some args & kwargs
            is_flow = True
            for keyword in function.decorator_list[0].keywords:
                # 'name' can be passed only as keyword in flow
                if keyword.arg == "name":
                    if sys.version_info.minor >= 8:
                        flow_name = keyword.value.value
                    else:
                        flow_name = keyword.value.s

                    break
        if is_flow:
            schedule = [
                schedule for schedule in schedules if schedule.lineno < function.lineno
            ]
            if schedule:
                schedule = schedule[0]
                schedules.remove(schedule)
            else:
                schedule = None
            tag = [tag for tag in tags if tag.lineno < function.lineno]
            if tag:
                tag = tag[0]
                tags.remove(tag)

            flow = FlowDataFromFile(
                name=flow_name,
                file_path=self.file_path,
                file_modified=os.path.getmtime(self.file_path)
                if self.file_path
                else None,
                schedule=Schedule(**schedule.dict(exclude={"lineno"}))
                if schedule
                else None,
                tags=tag.tags if tag else [],
                flow_data=FlowData(blob=self.file_data, encoding="text"),
            )
            self.flows.append(flow)



from fastapi import APIRouter
from fastflows.schemas.flow import FlowDeployInput, Flow
from fastflows.schemas.flow_run import FlowRunInput
from fastflows.core.flow import run_flow, list_flows, get_flow_runs_list, deploy_flows
from fastflows.routers import handle_rest_errors
from typing import List, Optional

router = APIRouter(prefix="/flows", tags=["flows"])


@router.post("", response_model=List[Flow])
@handle_rest_errors
async def create_flow_deployment(
    flow_deploy_input: Optional[FlowDeployInput] = FlowDeployInput(),
) -> List[Flow]:
    """if no input - deploy all flows from FLOWS_HOME"""
    # todo: should we separate routes? 1) create flow 2) create deployment
    return deploy_flows(flow_deploy_input)


@router.post("/{flow_id}")
async def init_flow_run(
    flow_id: str, flow_run_input: Optional[FlowRunInput] = FlowRunInput()
) -> Flow:
    """
    :param flow_id: Flow id in Prefect to run
    """
    return run_flow(flow_id, by_id=True, flow_run_input=flow_run_input)


@router.post("/name/{flow_name}")
async def init_flow_run_by_name(
    flow_name: str, flow_run_input: Optional[FlowRunInput] = FlowRunInput()
) -> Flow:
    """
    :param flow_name: Flow name in Prefect to run
    """
    return run_flow(flow_name, by_id=False, flow_run_input=flow_run_input)


@router.get("")
async def list_all_registered_flows_in_fast_flows():
    return list_flows()


@router.get("/{flow_id}/flow_runs")
async def list_all_flow_runs_by_flow_id(flow_id: str):
    """
    :param flow_id: Flow id in Prefect to run
    """
    return get_flow_runs_list(flow_id, by_id=True)


@router.post("/name/{flow_name}/flow_runs")
async def list_all_flow_runs_by_flow_name(flow_name: str):
    """
    :param flow_name: Flow id in Prefect to run
    """
    return get_flow_runs_list(flow_name, by_id=False)



class Singleton(type):
    # singleton implementation for Catalog and Providers
    _instance = {}

    def __call__(self, *args, **kwargs):
        if self not in self._instance:
            self._instance[self] = super().__call__(*args, **kwargs)
        return self._instance[self]



from pydantic import BaseModel, Field, root_validator
from enum import Enum
from fastflows.schemas.misc import get_hash_from_data, Schedule
from typing import Optional, List


# Flow Data from file or input
class Encodings(str, Enum):

    TEXT = "text"
    JSON = "json"
    CLOUDPICKLE = "cloudpickle"
    BLOCKSTORAGE = "blockstorage"


class ScheduleFromFile(Schedule):
    lineno: int


class FlowData(BaseModel):

    encoding: Encodings = Encodings.TEXT.value
    blob: str


class BaseFlowData(BaseModel):
    """Base class for flow data"""

    name: str
    tags: List[str] = Field(default_factory=list)


class FlowDataFromFile(BaseFlowData):
    """model to store information from Flow File"""

    file_path: Optional[str]
    flow_data: Optional[FlowData]
    # file_modified can be None if flow data from user input
    file_modified: Optional[int]
    schedule: Optional[Schedule]
    # hash from file/flow code (md5)
    deployment_name: Optional[str]

    @root_validator(pre=True)
    def _generate_deployment_name(
        cls, values: dict  # noqa: N805,B902 because of Pydantic
    ) -> dict:
        values["deployment_name"] = get_hash_from_data(values["flow_data"].blob)
        return values


class TagsFromFile(BaseModel):
    tags: List[str]
    lineno: int



import os


class FlowsStorageBase:
    def __init__(self, storage_path: str) -> None:
        self.storage_path = storage_path

    def list():
        raise NotImplementedError(
            "Flows Storage should implement list method to get all files in flows path directory."
        )

    def read():
        raise NotImplementedError(
            "Flows Storage should implement read method to get flow file content"
        )


class LocalStorage(FlowsStorageBase):
    def list(self):
        if not os.path.isdir(self.storage_path):
            raise ValueError(
                f"Flows Home must be a folder. You provided: {self.storage_path}"
            )
        return os.listdir(self.storage_path)

    def read(self):
        pass



from enum import Enum
from pydantic import BaseModel


class TaskState(Enum):

    # possible list of task states from Prefect
    PENDING = "pending"
    QUEUED = "queued"
    RETRYING = "retrying"
    RUNNING = "running"
    CACHED = "cached"
    FINISHED = "finished"
    LOOPED = "looped"
    SCHEDULED = "scheduled"
    SUBMITTED = "submitted"
    SKIPPED = "skipped"
    SUCCESS = "success"
    FAILED = "failed"
    MAPPED = "mapped"


class TaskRunState(BaseModel):
    id: str
    state: TaskState



from fastflows.schemas.flow_run import StateBase

from fastflows.providers import provider


def get_flow_run_details(flow_run_id: str):
    """
    :param flow_run_id: Flow Run Id in Prefect to get info about

    """
    return provider.get_flow_run_details(flow_run_id)


def update_flow_run_state(flow_run_id: str, state: StateBase):
    """
    :param flow_run_id: Flow Run Id in Prefect to get info about

    """
    return provider.update_flow_run_state(flow_run_id, state)



from fastapi import APIRouter
from fastflows.schemas.task_run import TaskRunState
from fastflows.core.task_run import update_task_run_state

router = APIRouter(prefix="/task-runs", tags=["task-runs"])


@router.patch("/{task_id}", response_model=TaskRunState)
async def update_task_run_state_route(task_run_id: str):
    """
    :param flow_run_id: Flow Run Id in Prefect
    :param task_id: Task Id to update

    """
    return update_task_run_state(task_run_id)



import os
import logging
from fastflows.errors import FlowNotFound
from fastflows.config.app import configuration as cfg
from fastflows.core.deployment import get_last_deployments_per_flow
from fastflows.core.catalog.storage import LocalStorage
from fastflows.core.catalog.reader import FlowFileReader
from fastflows.schemas.flow import Flow, PrefectFlowResponse, FlowDeployInput
from fastflows.schemas.flow_data import FlowDataFromFile, BaseFlowData
from fastflows.providers import provider
from fastflows.core.catalog.cache import CatalogCache
from fastflows.core.utils.singleton import Singleton
from fastflows.schemas.deployment import DeploymentSpec, FlowData, DeploymentResponse
from typing import List, Optional, Dict, Union


class Catalog(metaclass=Singleton):
    """main class to register all flows"""

    # because class is singleton - always will be one instance
    catalog: Dict[str, Flow] = {}
    catalog_by_id: Dict[str, Flow] = {}

    storages = {
        # add here API to work with flows on S3, or any other store
        "local": LocalStorage
    }

    def __init__(
        self,
        flows_home_path: str = cfg.FLOWS_HOME,
        storage_type: str = cfg.FLOWS_STORAGE_TYPE,
    ) -> None:

        self.flows_home_path = flows_home_path
        self.storage_type = storage_type
        self.cache = CatalogCache()
        self.get_catalog_from_cache()

    def _get_storage(self):
        storage = self.storages.get(self.storage_type)
        if storage is None:
            raise Exception(
                f"Flows storage type is not supported. Must be one of: {self.storages}"
            )
        return storage(self.flows_home_path)

    def get_flow_data(self, flow_file_path: str) -> FlowData:
        with open(flow_file_path) as f:
            return FlowData(blob=f.read())

    def get_catalog_from_cache(self):
        cached_data = self.cache.read()
        for flow_name, flow_data in cached_data.items():
            flow = Flow(**flow_data)
            self.catalog[flow_name] = flow
            self.catalog_by_id[flow_data["id"]] = flow

    def set_flows_path(self, new_path: str) -> None:
        # method to use in test
        self.flows_home_path = new_path

    def _blob_data_processing(self, flow_data: str) -> FlowData:

        flow_file = FlowFileReader(flow_data=flow_data)

        if not flow_file.is_flows:
            # file exists, but there is no flow inside
            raise FlowNotFound("Flow was not found in provided data")

        return flow_file.flows

    def _flow_path_processing(self, flow_path: str) -> str:

        # to have absolute path
        if self.flows_home_path not in flow_path:
            flow_path = os.path.join(self.flows_home_path, flow_path)

        if not os.path.exists(flow_path):
            # file does not exist
            raise FlowNotFound(f"Flow path '{flow_path}' does not exist'")

        flow_file = FlowFileReader(file_path=flow_path)

        if not flow_file.is_flows:
            # file exists, but there is no flow inside
            raise FlowNotFound(f"Flow was not found in file with path '{flow_path}''")

        return flow_file.flows

    def _get_flows_from_path(
        self, flow_input: FlowDeployInput
    ) -> List[FlowDataFromFile]:
        if flow_input.flow_path:
            flows_in_folder = self._flow_path_processing(flow_input.flow_path)
        elif flow_input.flow_data:
            flows_in_folder = self._blob_data_processing(flow_input.flow_data.blob)
        else:
            flows_in_folder = self.process_flows_folder()

        return flows_in_folder

    def _filter_flows_by_name(
        self,
        flows_in_folder: List[FlowDataFromFile],
        flow_name: str,
        flows_path: Optional[str],
    ) -> List[FlowDataFromFile]:
        flows_in_folder = [flow for flow in flows_in_folder if flow.name == flow_name]

        if len(flows_in_folder) == 0:
            if not flows_path:
                err_message = f"Flow with name '{flow_name}' was not found in flows home dir '{self.flows_home_path}'"
            else:
                err_message = (
                    f"Flow with name '{flow_name}' was not found in path '{flows_path}'"
                )
            raise FlowNotFound(err_message)
        return flows_in_folder

    def _clean_up_catalog_cache(self, flows_in_folder: List[FlowDataFromFile]) -> None:
        flow_names_in_folder = [flow.name for flow in flows_in_folder]

        clean_catalog = {
            flow_name: flow_data
            for flow_name, flow_data in self.catalog.items()
            if flow_name in flow_names_in_folder
        }
        self.catalog = clean_catalog

    def _get_version_from_tag(self, tags: List[str]) -> int:
        for tag in tags:
            if tag.startswith(cfg.VERSION_PREFIX):
                version = tag.split(cfg.TAG_DELIMITER)[1]
                return int(version)
        else:
            return 1

    def compare_cache_with_prefect(self):
        deployments_from_prefect = get_last_deployments_per_flow()
        flows_from_prefect = provider.get_flows()
        self.catalog = {}
        self.catalog_by_id = {}
        for flow_data in flows_from_prefect:
            _flow_data = flow_data.dict()
            last_flow_deployment = deployments_from_prefect.get(flow_data.id)
            if last_flow_deployment:
                _flow_data["deployment_id"] = last_flow_deployment.id
                _flow_data["deployment_name"] = last_flow_deployment.name
                _flow_data["version"] = self._get_version_from_tag(
                    last_flow_deployment.tags
                )
                flow = Flow(**_flow_data)
                self.catalog[_flow_data["name"]] = flow
                self.catalog_by_id[_flow_data["id"]] = flow
            else:
                logging.info(f"Flow without any deployment: {_flow_data['name']}")

    def register_and_deploy(
        self, flow_input: Optional[FlowDeployInput] = FlowDeployInput()
    ) -> List[Flow]:
        """
        flow_name - name of the Flow to deploy
        flow_path - path to the Flow to deploy

        if no flow_name & no flow_path - deploy all flows from the folder

        """
        logging.info("Checking for new flows or updates")

        # get flows from Prefect
        self.compare_cache_with_prefect()

        flows_updated = False

        flows_in_folder = self._get_flows_from_path(flow_input)

        if flow_input.flow_name:
            flows_in_folder: List[FlowDataFromFile] = self._filter_flows_by_name(
                flows_in_folder, flow_input.flow_name, flow_input.flow_path
            )

        if not flow_input.flow_path:
            # sync .fastflows cache with current state of FLOWS_HOME
            self._clean_up_catalog_cache(flows_in_folder)

        flows_deployed = []

        for flow_file in flows_in_folder:
            # if it will be deploy with flow_name - in flows_in_folder will be only one flow with that name
            flow: Flow = self._process_flow_file_deployment(flow_file, flow_input.force)
            if flow:
                flows_deployed.append(flow)
                flows_updated = True

        if not flows_updated:
            logging.info("No new flows or updates was found")

        self.cache.write(self.catalog)
        return flows_deployed

    def _get_deployment_version(
        self, flow: Union[Flow, PrefectFlowResponse], flow_file: FlowDataFromFile
    ) -> int:
        if not isinstance(flow, PrefectFlowResponse):
            version = flow.version
            if flow.deployment_name != flow_file.deployment_name:
                # mean code was changed
                version += 1
        else:
            version = 1
        return version

    def _process_flow_file_deployment(
        self, flow_file: FlowDataFromFile, force: bool
    ) -> Optional[Flow]:

        flow: Union[Flow, PrefectFlowResponse] = self._prepare_flow_for_deployment(
            flow_file, force
        )

        if not flow:
            # mean no updates - no need to redeploy
            return

        version = self._get_deployment_version(flow, flow_file)

        deployment = self._deploy_flow(flow, flow_file, version)

        flow = Flow(
            id=flow.id,
            name=flow_file.name,
            file_path=flow_file.file_path,
            deployment_id=deployment.id,
            deployment_name=deployment.name,
            version=version,
        )
        self.catalog[flow_file.name] = flow
        self.catalog_by_id[flow.id] = flow
        return flow

    def _prepare_flow_for_deployment(
        self, flow_file: FlowDataFromFile, force: bool
    ) -> Optional[PrefectFlowResponse]:
        if flow_file.name in self.catalog:
            if (
                flow_file.deployment_name
                and flow_file.file_modified
                == self.catalog[flow_file.name].deployment_name
            ) and not force:
                # no updates in file, no need to re-deploy
                return None
            logging.info(f"Deploying new version of flow '{flow_file.name}' in Prefect")
            flow = self.catalog[flow_file.name]
        else:
            logging.info(f"Registering new flow '{flow_file.name}' in Prefect")
            flow: PrefectFlowResponse = provider.create_flow(
                BaseFlowData(**flow_file.dict())
            )
        return flow

    def _deploy_flow(
        self, flow: Flow, flow_file: FlowDataFromFile, version: int
    ) -> DeploymentResponse:
        deployment_spec = DeploymentSpec(
            flow_id=flow.id,
            flow_data=flow_file.flow_data,
            # todo: add parsing for schedule
            deployment_name=flow_file.name,
            schedule=None,
            is_schedule_active=False,
            version=version,
        )
        deployment_response = provider.deploy_flow(deployment_spec)
        return deployment_response

    def _get_full_flow_location(self, flow_file_name: str) -> str:
        return os.path.join(self.flows_home_path, flow_file_name)

    def process_flows_folder(self) -> List[FlowDataFromFile]:
        """list flows from FLOWS_HOME without register them or load them to Prefect if 'register' True"""

        flows_in_storage = self._get_storage().list()

        flows_in_folder = []

        for file_name in flows_in_storage:
            if not file_name.endswith(".py"):
                continue
            full_flow_path = self._get_full_flow_location(file_name)

            # get data from flow file
            flow_file = FlowFileReader(full_flow_path)
            flows_in_folder += flow_file.flows
        return flows_in_folder

    def list_flows(self) -> List[str]:
        return list(catalog.items())


catalog = Catalog.catalog
catalog_by_id = Catalog.catalog_by_id



from fastflows.providers import provider
from typing import Dict
from fastflows.schemas.deployment import DeploymentResponse


def get_last_deployments_per_flow() -> Dict[str, DeploymentResponse]:
    deployments_from_prefect = provider.get_deployments()

    deployments_by_flow_id = {}

    for deployment in deployments_from_prefect:
        deployments_by_flow_id[deployment.flow_id] = deployment
    return deployments_by_flow_id



"""App exceptions module."""
from fastapi import Request
from starlette.responses import JSONResponse


class AppExceptionError(Exception):
    """Application exception base error class."""

    def __init__(self, status_code: int, error: str, context: dict):
        """Handle application exceptions initialization."""
        self.exception_case = self.__class__.__name__
        self.status_code = status_code
        self.error = error
        self.context = context

    def __str__(self):
        """Define representation of application exception instance."""
        return (
            f"<AppExceptionError {self.exception_case} - error={self.error}"
            + f"status_code={self.status_code} - context={self.context}>"
        )


async def app_exception_handler(request: Request, exc: AppExceptionError):
    """Handle json representation of application exception."""
    return JSONResponse(
        status_code=exc.status_code,
        content={
            "error": exc.error,
            "context": exc.context,
        },
    )


class AppException:
    """Application exception class."""

    pass



from pydantic import BaseModel


class FastFlowException(Exception):
    def __init__(self, message: str = "Error while processing request"):

        super().__init__(message)


class FlowNotFound(FastFlowException):
    def __init__(self, message: str = "Flow was not found"):
        super().__init__(message)


def api_response_handler(response_model: BaseModel, message: str):
    def decorator(func):
        def wrapper(*args, **kwargs):
            try:
                response = func(*args, **kwargs)
                if response.status_code != 200:
                    response = response.json()
                    return response_model(**response)
                else:
                    raise FastFlowException()
                return func(*args, **kwargs)
            except Exception as e:
                raise FastFlowException(message) from e

        return wrapper

    return decorator



import os
from fastflows.config.app import configuration as cfg
from ast import literal_eval


class CatalogCache:

    separator = " >> "

    def __init__(self):
        self.data = self.read()

    def check_or_create_home_folder(self) -> None:

        if not os.path.exists(cfg.FLOWS_HOME):
            os.makedirs(cfg.FLOWS_HOME, exist_ok=True)

    def read(self) -> None:
        self.check_or_create_home_folder()
        try:
            with open(cfg.FASTFLOWS_CATALOG_CACHE, "r") as f:
                raw_data = f.read()
        except IOError:
            with open(cfg.FASTFLOWS_CATALOG_CACHE, "w+") as f:
                f.write("")
                raw_data = ""

        self.data = {}
        if raw_data:
            # get list with flow_name, flow_data
            data = raw_data.split("\n")
            flows_data_list = []
            for line in data:
                line = line.split(self.separator)
                if len(line) > 1:
                    flows_data_list.append(line[0])
                    flows_data_list.append(literal_eval(line[1]))

            self.data = dict(zip(flows_data_list[0::2], flows_data_list[1::2]))
        return self.data

    def write(self, catalog_data: dict) -> None:
        data = self.catalog_dict_to_cache_str(catalog_data)

        with open(cfg.FASTFLOWS_CATALOG_CACHE, "w") as f:
            f.write(data)

    def catalog_dict_to_cache_str(self, data: dict) -> str:
        cache_str = ""
        for key, value in data.items():
            value = value.dict()
            cache_str += f"{key}{self.separator}{value}\n"
        return cache_str



from typing import List
from fastflows.schemas.flow import Flow, FlowDeployInput
from fastflows.schemas.flow_run import FlowRunResponse, FlowRunInput
from fastflows.errors import FlowNotFound
from fastflows.core.catalog.catalog import catalog, Catalog, catalog_by_id
from fastflows.providers import provider


def get_flow_id(flow_name: str, by_id: bool):
    flow_id = flow_name if by_id else get_flow_by_name(flow_name).id
    return flow_id


def get_flow_by_name(flow_name: str) -> Flow:
    """
        get flow from Catalog, if flow does not exists - try to re-register
    flows & try one more time
    """

    flow = catalog.get(flow_name)

    if not flow:
        flow = register_flow_and_check(flow_name, catalog=catalog)
    return flow


def register_flow_and_check(flow_id: str, catalog: dict) -> Flow:
    # try to register flows, maybe they was updated during runtime
    Catalog().register_and_deploy(FlowDeployInput())
    flow = catalog.get(flow_id)
    if not flow:
        raise FlowNotFound(
            f"Flow was not found in Catalog. Available flows {list(catalog.keys())}"
        )
    return flow


def get_flow_by_id(flow_id: str) -> Flow:
    flow = catalog_by_id.get(flow_id)
    if not flow:
        flow = register_flow_and_check(flow_id, catalog=catalog_by_id)
    return flow


def run_flow(
    flow_name: str, by_id: bool, flow_run_input: FlowRunInput
) -> FlowRunResponse:
    """run flow by name or id"""
    flow_id = get_flow_id(flow_name, by_id)
    flow = get_flow_by_id(flow_id)
    if flow:
        return provider.run_flow(
            deployment_id=flow.deployment_id,
            flow_run_params=flow_run_input,
        )
    else:
        err_message = f"Flow with {'ID' if by_id else 'name'} {flow_name} was not found"
        raise FlowNotFound(err_message)


def get_flow_runs_list(flow_name: str, by_id: bool) -> List[FlowRunResponse]:
    flow_id = get_flow_id(flow_name, by_id)
    return provider.list_flow_runs(flow_id)


def list_flows():
    # flows that already was registered in Prefect
    return catalog


def list_flows_from_file_home(flow_path: str) -> List[Flow]:
    # they cannot be registered in Prefect, just list from FLOWS_HOME
    return list(Catalog(flow_path).process_flows())


def deploy_flows(
    flow_input: FlowDeployInput,
) -> List[Flow]:
    return Catalog(flows_home_path=flow_input.flows_home_path).register_and_deploy(
        flow_input
    )



"""Request exceptions module."""
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import RequestValidationError
from starlette.exceptions import HTTPException
from starlette.requests import Request
from starlette.responses import JSONResponse
from starlette.status import HTTP_422_UNPROCESSABLE_ENTITY


async def http_exception_handler(request: Request, exc: HTTPException) -> JSONResponse:
    """Handle HTTP exception result."""
    return JSONResponse({"detail": exc.detail}, status_code=exc.status_code)


async def request_validation_exception_handler(
    request: Request, exc: RequestValidationError
) -> JSONResponse:
    """Handle validation exception result."""
    return JSONResponse(
        status_code=HTTP_422_UNPROCESSABLE_ENTITY,
        content={"detail": jsonable_encoder(exc.errors())},
    )



from fastflows.schemas.task_run import TaskRunState, TaskState


def update_task_run_state(task_run_id: str):
    # call provider API
    return TaskRunState(id=task_run_id, state=TaskState.MAPPED)


def get_task_run_state(task_run_id: str):
    pass



"""Authn and Authz module."""
from fastflows.config.app import configuration as cfg
from fastapi_opa import OPAConfig
from fastapi_opa.auth import OIDCAuthentication
from fastapi_opa.auth import OIDCConfig


# The hostname of your Open Policy Agent instance
opa_host = cfg.OPA_URL
if opa_host:
    # In this example we use OIDC authentication flow (using Keycloak)
    oidc_config = OIDCConfig(
        well_known_endpoint=cfg.OIDC_WELL_KNOWN_ENDPOINT,
        # well known endpoint
        app_uri=cfg.HOST,  # host where this app is running
        # client id of your app configured in the identity provider
        client_id=cfg.OIDC_CLIENT_ID,
        # the client secret retrieved from your identity provider
        client_secret=cfg.OIDC_CLIENT_SECRET,
    )
    oidc_auth = OIDCAuthentication(oidc_config)
    opa_config = OPAConfig(authentication=oidc_auth, opa_host=opa_host)
else:
    opa_config = None



import datetime
from typing import List, Optional
from pydantic import BaseModel, Field
from fastflows.schemas.deployment import DeploymentInputParams
from fastflows.schemas.flow_data import FlowData
from fastflows.config.app import configuration as cfg


# before prefect call
class FlowDeployInput(BaseModel):
    """REST Create flow Input model"""

    flows_home_path: Optional[str] = cfg.FLOWS_HOME
    flow_data: Optional[FlowData]
    flow_name: Optional[str]
    flow_path: Optional[str]
    deployment_params: Optional[DeploymentInputParams]
    force: bool = Field(False, description="Force deploy all flows")


# Flow Data after communication with Prefect
class Flow(BaseModel):
    """Created in Prefect Flow with deployment"""

    id: str
    name: str
    file_path: Optional[str]
    deployment_id: str
    deployment_name: str
    version: int


class PrefectFlowResponse(BaseModel):

    name: str
    tags: List[str] = Field(default_factory=list)
    id: str
    created: datetime.datetime
    updated: datetime.datetime



"""Main module."""
from typing import Any
import logging
import loguru
import uvicorn
from time import sleep
from fastflows.config.app import configuration as cfg
from fastflows.config.auth import opa_config
from fastflows.config.logging import create_logger
from fastflows.utils.app_exceptions import app_exception_handler
from fastflows.utils.app_exceptions import AppExceptionError
from fastflows.utils.request_exceptions import http_exception_handler
from fastflows.utils.request_exceptions import request_validation_exception_handler
from fastflows.routers import flows, flow_runs
from fastflows.errors import FastFlowException

from fastflows.core.catalog import Catalog
from fastapi import FastAPI
from fastapi.exceptions import RequestValidationError
from fastapi_opa import OPAMiddleware
from mangum import Mangum
from starlette.exceptions import HTTPException as StarletteHTTPException
from starlette.middleware.cors import CORSMiddleware
from fastflows.providers import provider


class FastFlowsAPI(FastAPI):
    """Subclass of FastAPI that possesses a logger attribute."""

    def __init__(self, **extra: Any):
        """Included the self.logger attribute."""
        super().__init__(**extra)
        self.logger: loguru.Logger = loguru.logger


def create_app() -> FastFlowsAPI:
    """Handle application creation."""
    app = FastFlowsAPI(title="FastFlowsApi", root_path=cfg.ROOT_PATH, debug=True)

    # Set all CORS enabled origins
    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    @app.exception_handler(StarletteHTTPException)
    async def custom_http_exception_handler(request, e):
        return await http_exception_handler(request, e)

    @app.exception_handler(RequestValidationError)
    async def custom_validation_exception_handler(request, e):
        return await request_validation_exception_handler(request, e)

    @app.exception_handler(AppExceptionError)
    async def custom_app_exception_handler(request, e):
        return await app_exception_handler(request, e)

    if cfg.OPA_URL:
        app.add_middleware(OPAMiddleware, config=opa_config)

    app.logger = create_logger(name="app.main")

    return app


app = create_app()


@app.on_event("startup")
async def startup_event():
    provider.healthcheck()
    # todo: improve to avoid re-deploying each time when server up & run
    if cfg.FASTFLOWS_AUTO_DEPLOYMENT == 1:
        logging.info("Register Flows in Prefect")
        try:
            Catalog().register_and_deploy()
        except FastFlowException:
            # error during connection to DB, maybe problem at the prefect start
            # let's wait & try one more time
            sleep(10)
            Catalog().register_and_deploy()


app.include_router(flows.router)
app.include_router(flow_runs.router)


def app_run():
    uvicorn.run(
        "fastflows.main:app",
        port=cfg.FASTFLOWS_PORT,
        reload=cfg.RELOAD,
        access_log=cfg.ACCESS_LOG,
    )


if cfg.AWS_LAMBDA_DEPLOY:
    # to make it work with Amazon Lambda,
    # we create a handler object
    handler = Mangum(app)

if __name__ == "__main__":
    app_run()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fastflows',
 'fastflows.config',
 'fastflows.core',
 'fastflows.core.catalog',
 'fastflows.core.utils',
 'fastflows.providers',
 'fastflows.routers',
 'fastflows.schemas',
 'fastflows.utils']

package_data = \
{'': ['*']}

install_requires = \
['fastapi-opa>=1.3.7,<2.0.0',
 'httpx>=0.23.0,<0.24.0',
 'loguru>=0.6.0,<0.7.0',
 'mangum>=0.15.0,<0.16.0',
 'pydantic[dotenv]>=1.9.1,<2.0.0',
 'rich>=12.4.4,<13.0.0',
 'typer>=0.4.1,<0.5.0',
 'uvicorn>=0.17.6,<0.18.0']

entry_points = \
{'console_scripts': ['fastflows = fastflows.cli:app']}

setup_kwargs = {
    'name': 'fastflows',
    'version': '0.1.0.dev1657133810',
    'description': 'FastFlows is a FastAPI server & command line tool to comunicate with Prefect 2.0 as a Workflow manager to deploy, run, track flows and more.',
    'long_description': "## Run FastFlow server\n\nFastFlows is a FastAPI server & command line tool to comunicate with Prefect 2.0 as a Workflow manager to deploy, run, track flows and more.\n\nTo start work with FastFlows you should define at least 2 environment variables:\n\n```console\n\n    # Prefect API Server address\n    PREFECT_URI=http://localhost:4200\n\n    # Path to folder with flows\n    FLOWS_HOME=flows\n\n```\n\nIf you want to define variables with env prefix, for example, like 'LOCAL_PREFECT_URI' or 'DEV_PREFECT_URI' you can use environment variable 'ENV_NAME'\n\nIf Fastflow will see 'ENV_NAME' variable in environment - it will search for variables with prefix defined in this ENV_NAME, for example:\n\nif ENV_NAME = 'LOCAL'\n\nFastflows will read variables like LOCAL_PREFECT_URI and LOCAL_FLOWS_HOME,\n\nif ENV_NAME = 'dev', then fastflow will expect variables like 'dev_PREFECT_URI' and 'dev_FLOWS_HOME'\n\n## Build FastFlows\n\nTo build stand alone image:\n\n```console\n\n    docker build . -f docker/Dockerfile -t fastflows\n\n```\n\n### Run Prefect witn DB in Docker-Compose\n\n```console\n\n    docker-compose -f ./docker/docker-compose.yml up  --build\n\n```\n\nTo enter UI:\n\n```console\n    # if you will try to use 0.0.0.0 you will not see any data because of CORS issues\n    http://localhost:4200/flows\n\n```\n\n### Run cli\n\n```console\n\n    fastflows --help\n\n```\n\n### Flows Deployment\n\n#### Auto deployment\n\nDeployment of Flows can be done by FastFlows automatically: if there is a new flow or changes in FLOWS_HOME directory - FastFlow create new deployment. To disable auto deployment set env variable to 0\n\n```console\n\n    FASTFLOWS_AUTO_DEPLOYMENT = 0\n\n```\n",
    'author': 'Francesco Bartoli',
    'author_email': 'francesco.bartoli@geobeyond.it',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/geobeyond/fastflows',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import httpx
from fastflows.config.app import configuration as cfg
from fastflows.providers.base import BaseProvider
from fastflows.schemas.deployment import DeploymentSpec, DeploymentResponse
from fastflows.schemas.flow import PrefectFlowResponse
from fastflows.schemas.flow_data import BaseFlowData
from fastflows.providers.utils import api_response_handler
from fastflows.schemas.flow_run import (
    FlowRunResponse,
    InitFlowRun,
    StateBase,
    UpdateStateResponse,
)
from typing import List, Optional, Dict


class PrefectProvider(BaseProvider):

    uri = f"{cfg.PREFECT_URI}/api"

    def __init__(self) -> None:
        self.client = httpx.Client(timeout=cfg.PREFECT_API_TIMEOUT)

    @api_response_handler(
        message="Cannot run the flow with flow_id {}",
    )
    def run_flow(
        self, deployment_id: str, flow_run_params: InitFlowRun
    ) -> FlowRunResponse:

        return self.client.post(
            f"{self.uri}/deployments/{deployment_id}/create_flow_run",
            json=flow_run_params.dict(),
        )

    @api_response_handler(...)
    def approve_task(self, flow_run_id: str) -> None:
        """need to implement manual_approve state firstly in Prefect"""
        # response = self.client.post(f"{self.uri}/flow_run_states/{flow_run_id}")
        # return response
        raise NotImplementedError()

    @api_response_handler(
        message="Error while deploying flow. Description:",
    )
    def get_task_run(self, task_run_id: str) -> None:
        """https://orion-docs.prefect.io/api-ref/rest-api/#/Task%20Runs/read_task_run_task_runs__id__get"""

        response = self.client.post(f"{self.uri}/task_runs/{task_run_id}/")
        return response

    @api_response_handler(
        message="Error while deleting deployment by_id",
    )
    def deploy_flow(self, deployment: DeploymentSpec) -> DeploymentResponse:
        """
        call deployment API:
            https://orion-docs.prefect.io/api-ref/rest-api/#/Deployments/create_deployment_deployments__post

            :param deployment: Deployment configuration

        """
        response = self.client.post(
            f"{self.uri}/deployments/", json=deployment.dict(exclude={"version"})
        )
        return response

    @api_response_handler(
        message="Error while creating new Flow.",
    )
    def create_flow(self, flow_request: BaseFlowData) -> PrefectFlowResponse:
        response = self.client.post(f"{self.uri}/flows/", json=flow_request.dict())
        return response

    @api_response_handler(
        message="Error during getting all flows from Prefect",
        response_model=List[PrefectFlowResponse],
    )
    def get_flows(self, filters: Optional[Dict] = None) -> List[PrefectFlowResponse]:
        response = self.client.post(f"{self.uri}/flows/filter", json=filters or {})
        return response

    @api_response_handler(
        message="Error during getting all deployments from Prefect",
        response_model=List[DeploymentResponse],
    )
    def get_deployments(
        self, filters: Optional[Dict] = None
    ) -> List[DeploymentResponse]:
        response = self.client.post(
            f"{self.uri}/deployments/filter", json=filters or {}
        )
        return response

    @api_response_handler(
        message="Error while getting Flow Runs by flow_id.",
        response_model=List[FlowRunResponse],
    )
    def list_flow_runs(self, flow_id: str) -> List[FlowRunResponse]:
        payload = {"flows": {"id": {"any_": [flow_id]}}}
        response = self.client.post(
            f"{self.uri}/flow_runs/filter",
            json=payload,
        )
        return response

    @api_response_handler(
        message="Error while getting FlowRun details.",
        append_api_error=True,
        response_model=List[FlowRunResponse],
    )
    def get_flow_run_details(self, flow_run_id: str) -> List[FlowRunResponse]:
        response = self.client.get(f"{self.uri}/flow_runs/{flow_run_id}")
        return response

    @api_response_handler(
        message=f"Prefect does not answer on Healthcheck. Looks like Prefect server is unavailable on address {cfg.PREFECT_URI}",
    )
    def healthcheck(self) -> str:
        response = self.client.get(f"{self.uri}/hello")
        return response

    @api_response_handler(
        message="Error while updating FlowRun state.",
    )
    def update_flow_run_state(
        self, flow_run_id: str, state: StateBase
    ) -> UpdateStateResponse:
        response = self.client.post(
            f"{self.uri}/flow_runs/{flow_run_id}/set_state",
            json={"state": state.dict()},
        )
        return response



import logging
from typing import Optional, _GenericAlias, Callable
from pydantic import BaseModel
import fastflows
from fastflows import errors
from fastflows.schemas.misc import DefaultAPIResponseModel


def get_response_model(func: Callable, response_model: Optional[BaseModel]):

    if not response_model:
        if func.__annotations__.get("return"):
            if func.__annotations__["return"] == str:
                return None
            return func.__annotations__["return"]
        else:
            return DefaultAPIResponseModel
    else:
        _response_model = response_model

        if isinstance(_response_model, _GenericAlias):
            _response_model = _response_model.__args__[0]

        return _response_model


def api_response_handler(
    response_model: Optional[BaseModel] = None,
    message: str = "Some unexpected error occures during the API call",
    args_in_message: bool = False,
    append_api_error: bool = True,
):
    def decorator(func):
        _response_model = get_response_model(func, response_model)

        def wrapper(*args, **kwargs):
            try:
                response = func(*args, **kwargs)
                status_code_check = str(response.status_code).startswith
                if status_code_check("2"):
                    response = response.json()
                    if _response_model:
                        if isinstance(response, list):
                            processed_response = []
                            for item in response:
                                processed_response.append(_response_model(**item))
                        else:
                            processed_response = _response_model(**response)
                        return processed_response
                    else:
                        return response
                else:
                    raise fastflows.errors.FastFlowException(
                        f"Error API response: {response.text}. Status code: {response.status_code}"
                    )
            except Exception as e:

                logging.error(e)

                err_msg = (
                    message.format(args[1:] if len(args) > 1 else kwargs)
                    if args_in_message
                    else message
                )
                if append_api_error:
                    err_msg += f" {e.args[0]}"
                raise errors.FastFlowException(err_msg)

        return wrapper

    return decorator



"""Service result module."""
import inspect

from app.config.logging import create_logger
from app.utils.app_exceptions import AppExceptionError


logger = create_logger("app.utils.service_result")


class ServiceResult:
    """Service result class."""

    def __init__(self, arg):
        """Initialize the service result class."""
        if isinstance(arg, AppExceptionError):
            self.success = False
            self.exception_case = arg.exception_case
            self.status_code = arg.status_code
        else:
            self.success = True
            self.exception_case = None
            self.status_code = None
        self.value = arg

    def __str__(self):
        """Handle class identity."""
        if self.success:
            return "[Success]"
        return f'[Exception] "{self.exception_case}"'

    def __repr__(self):
        """Handle class representation."""
        if self.success:
            return "<ServiceResult Success>"
        return f"<ServiceResult AppException {self.exception_case}>"

    def __enter__(self):
        """Handle runtime instantiation."""
        return self.value

    def __exit__(self, *kwargs):
        """Handle runtime exit."""
        pass


def caller_info() -> str:
    """Handle information for the caller."""
    info = inspect.getframeinfo(inspect.stack()[2][0])
    return f"{info.filename}:{info.function}:{info.lineno}"


def handle_result(result: ServiceResult):
    """Handle result of a service."""
    if not result.success:
        with result as exception:
            logger.error(f"{exception} | caller={caller_info()}")
            raise exception
    with result as result:
        return result



# prefect-doc https://orion-docs.prefect.io/concepts/deployments/
import datetime
from enum import Enum
import uuid
from time import time
from pydantic import Field, validator, BaseModel, root_validator
from typing import List, Optional
from fastflows.schemas.flow_data import FlowData
from fastflows.schemas.misc import get_hash_from_data, Schedule
from fastflows.config.app import configuration as cfg


class FlowRunnerTypes(Enum):
    # possible variants from Prefect
    UNIVERSAL = "universal"
    DOCKER = "docker"
    KUBERNETES = "kubernetes"
    SUBPROCESS = "subprocess"


class FlowRunner(BaseModel):

    # config - is a set of arguments that you can send in chosen FlowRunner class (type)
    config: dict = {}
    type: FlowRunnerTypes = cfg.DEFAULT_FLOW_RUNNER_TYPE


class DeploymentInputParams(BaseModel):
    """deployment params that can be in input by user"""

    # can be input from rest api
    flow_data: Optional[FlowData]
    # properties what can be input by user with cli or REST API
    schedule: Optional[Schedule] = None
    is_schedule_active: bool = True
    parameters: Optional[dict] = {}
    version: Optional[int]
    tags: List[str] = Field(default_factory=list)


def generate_deployment_name():
    return datetime.datetime.now().isoformat()


class DeploymentSpec(DeploymentInputParams):

    name: Optional[str] = None
    flow_id: str = Field(default_factory=lambda: str(uuid.uuid4()))
    flow_runner: FlowRunner = Field(default_factory=lambda: FlowRunner())

    @validator("flow_id", "name")
    def not_empty_string(cls, value):
        if value.strip() == "":
            raise ValueError("Value cannot be empty string")
        return value

    @root_validator(pre=True)
    def generate_tags_and_name(cls, values) -> List[dict]:
        if not values.get("version"):
            values["version"] = 1

        if not values.get("tags"):
            values["tags"] = []

        values["tags"].extend(
            [
                f'{cfg.VERSION_PREFIX}{cfg.TAG_DELIMITER}{values["version"]}',
                f'ts:{str(time()).split(".")[0]}',
            ]
        )

        if not values.get("name"):
            values["name"] = get_hash_from_data(values["flow_data"].blob)

        return values


class DeploymentResponse(DeploymentSpec):

    id: str
    created: datetime.datetime
    updated: datetime.datetime



import os
from typing import Optional
from pydantic import BaseSettings


class Config(BaseSettings):

    ENV_NAME: Optional[str] = os.environ.get("ENV_NAME")
    # each environment read variables with own prefix
    # to override settings use environment variables
    # with prefix from 'env_prefix' in config
    PREFECT_URI: str = "http://localhost:8080"
    PREFECT_API_TIMEOUT: int = 120
    DEFAULT_FLOW_RUNNER_TYPE: str = "subprocess"

    # config for tags information & parsing flows properties
    VERSION_PREFIX: str = "ver"
    TAG_DELIMITER: str = ":"
    SCHEDULE_PROPERTY: str = "schedule"
    TAGS_PROPERTY: str = "tags"

    # path to search flows
    FLOWS_HOME: str = "flows/"
    # type of flows home directory: local, s3, etc. 'local' only supported from start
    FLOWS_STORAGE_TYPE: str = "local"
    FASTFLOWS_CATALOG_CACHE: str = "./flows/.fastflows"
    FASTFLOWS_AUTO_DEPLOYMENT: int = 1
    FASTFLOW_DEBUG: int = 1

    # logging
    LOG_LEVEL: str = "DEBUG"
    LOG_PATH: str = "/tmp"
    LOG_FILENAME: str = "fastflows.log"
    LOG_ENQUEUE: bool = True
    LOG_ROTATION: str = "1 days"
    LOG_RETENTION: str = "1 months"
    LOG_FORMAT: str = "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | <level>{level: <8}</level> | <cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> [id:{extra[request_id]}] - <level>{message}</level>"

    # uvicorn setting
    HOST: str = "0.0.0.0"
    FASTFLOWS_PORT: int = 5010
    RELOAD: bool = True
    ACCESS_LOG: bool = False
    ROOT_PATH: str = ""

    # auth
    # for test purposes values taken from tests in official repo:
    # https://github.com/busykoala/fastapi-opa/blob/a456f4e6f8f6cb90ca386bbcd5909af3ea44d646/tests/utils.py#L63
    OPA_URL: Optional[str] = ""
    OPA_HOST: Optional[str] = "0.0.0.0"
    OIDC_WELL_KNOWN_ENDPOINT: Optional[str] = "0.0.0.0/.well-known/openid-configuration"
    OIDC_CLIENT_ID: Optional[str] = "example-client"
    OIDC_CLIENT_SECRET: Optional[str] = "secret"

    # env settings
    AWS_LAMBDA_DEPLOY: bool = False

    class Config:
        env_file: str = ".env"
        env_prefix: str = (
            f'{os.environ.get("ENV_NAME")}_' if os.environ.get("ENV_NAME") else ""
        )


configuration = Config()
